/*
 * Copyright 2021 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.tomee.vestige.launcher;

import java.io.Closeable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLDecoder;
import java.net.URLStreamHandler;
import java.security.Policy;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.regex.Matcher;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.catalina.LifecycleException;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.webresources.StandardRoot;
import org.apache.commons.io.IOUtils;
import org.apache.openejb.AppModuleCreatorData;
import org.apache.openejb.OpenEJBCoreController;
import org.apache.openejb.OpenEJBException;
import org.apache.openejb.WebModuleCreatorData;
import org.apache.openejb.classloader.WebAppEnricher;
import org.apache.openejb.config.AppModule;
import org.apache.openejb.config.DeploymentLoader;
import org.apache.openejb.config.DeploymentModule;
import org.apache.openejb.config.WebModule;
import org.apache.openejb.jee.Application;
import org.apache.openejb.loader.Files;
import org.apache.openejb.loader.OpenEJBLoaderController;
import org.apache.openejb.loader.SystemInstance;
import org.apache.openejb.sxc.ApplicationXml;
import org.apache.tomcat.vestige.launcher.DelegateClassLoader;
import org.apache.tomcat.vestige.launcher.VestigeTomcatController;
import org.apache.tomcat.vestige.launcher.VestigeWar;
import org.apache.tomcat.vestige.launcher.webresources.VestigeJarEntryConnection;
import org.apache.tomcat.vestige.launcher.webresources.VestigeJarResourceSet;
import org.apache.tomee.catalina.CatalinaController;
import org.apache.xbean.finder.ResourceFinder;

import fr.gaellalire.vestige.core.VestigeCoreURLStreamHandler;
import fr.gaellalire.vestige.spi.resolver.AttachedClassLoader;
import fr.gaellalire.vestige.spi.resolver.VestigeJar;
import fr.gaellalire.vestige.spi.resolver.VestigeJarEntry;

/**
 * @author Gael Lalire
 */
public class VestigeTomEEController extends VestigeTomcatController implements OpenEJBCoreController, OpenEJBLoaderController, CatalinaController {

    private static final ThreadLocal<VestigeEar> VESTIGE_EAR_THREAD_LOCAL = new ThreadLocal<VestigeEar>();

    private static final ThreadLocal<VestigeWar> VESTIGE_WAR_THREAD_LOCAL = new ThreadLocal<VestigeWar>();

    private TomEEVestigeSystem tomEEVestigeSystem;

    private File conf;

    private File data;

    public VestigeTomEEController(final TomEEVestigeSystem tomEEVestigeSystem, final File conf, final File data) {
        this.tomEEVestigeSystem = tomEEVestigeSystem;
        this.conf = conf;
        this.data = data;
    }

    @Override
    public void warDeployed() {
        super.warDeployed();
        VESTIGE_WAR_THREAD_LOCAL.remove();
        VESTIGE_WAR_THREAD_LOCAL.remove();
    }

    @Override
    public void warUndeployed(final Closeable closeable) {
        super.warUndeployed(closeable);
        // TODO detach ear / ejb / rar / ear libs
    }

    private Map<Object, VestigeWar> map = new WeakHashMap<>();

    public void setResources(final StandardContext standardContext) {
        final VestigeWar vestigeWar = map.get(standardContext);

        standardContext.setResources(new StandardRoot() {

            @Override
            public void processWebInfLib() throws LifecycleException {
                super.processWebInfLib();
                if (vestigeWar != null) {
                    for (VestigeJar vestigeJar : vestigeWar.getVestigeWarDependencies()) {
                        addClassResources(new VestigeJarResourceSet(this, "/WEB-INF/classes", vestigeJar, Collections.<VestigeJar> emptyList(), "/"));
                    }
                }
            }
        });
    }

    public String removeExtension(final String name) {
        if (name.endsWith(".vear") || name.endsWith(".vwar")) {
            return name.replaceFirst(".....$", "");
        }
        return null;
    }

    public Class<? extends DeploymentModule> getModuleClass(final File jarFile) {
        if (jarFile.getName().endsWith(".vear")) {
            return AppModule.class;
        } else if (jarFile.getName().endsWith(".vwar")) {
            return WebModule.class;
        }
        return null;
    }

    public static void unzipExploded(final VestigeJar vestigeJar, final File destinationDir) throws IOException {
        for (VestigeJarEntry entry : Collections.list(vestigeJar.getEntries())) {
            String name = entry.getName();
            if (name.equals("META-INF/exploded-assembly.properties") || name.startsWith("META-INF/exploded-assembly-files/")) {
                continue;
            }
            File file = new File(destinationDir, name.replace('/', File.separatorChar));
            if (entry.isDirectory()) {
                file.mkdirs();
                continue;
            }
            file.getParentFile().mkdirs();
            try (FileOutputStream fos = new FileOutputStream(file)) {
                try (InputStream open = entry.open()) {
                    IOUtils.copy(open, fos);
                }
            }
        }
    }

    public boolean extract(final File file, final File destinationDir) throws IOException {
        if (file.getName().endsWith(".vear")) {
            final VestigeEar vestigeEar = VestigeEar.create(file, file.getName().substring(0, file.getName().length() - ".vear".length()));
            VESTIGE_EAR_THREAD_LOCAL.set(vestigeEar);
            VestigeJar vestigeEarJar = vestigeEar.getVestigeEar();
            try {
                unzipExploded(vestigeEarJar, destinationDir);
            } catch (final IOException e) {
                // If something went wrong, delete extracted dir to keep things clean
                Files.delete(destinationDir);
                throw e;
            }
            for (WarInEar warInEar : vestigeEar.getWarInEars()) {
                // create a vwar
                try {
                    String entryName = warInEar.getEntryName();
                    if (entryName.endsWith(".war")) {
                        entryName = entryName.substring(0, entryName.length() - ".war".length());
                    }
                    VWarGenerator.generate(warInEar, new File(destinationDir, entryName.replace('/', File.separatorChar) + ".vwar"));
                } catch (Exception e) {
                    throw new IOException("Unable to create vwar", e);
                }
            }
            List<? extends VestigeJar> vestigeEarDependencies = vestigeEar.getVestigeEarDependencies();
            for (VestigeJar vestigeJar : vestigeEarDependencies) {
                JarMetaData jarMetaData = vestigeEar.getJarMetaDataByURL().get(vestigeJar.getCodeBase().toString());
                Map<String, String> overrideByEntryName = jarMetaData.getOverrideByEntryName();
                File outputFile = new File(destinationDir, jarMetaData.getEarEntryName().replace('/', File.separatorChar));
                if (overrideByEntryName != null) {
                    try (ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(outputFile))) {
                        for (VestigeJarEntry entry : Collections.list(vestigeJar.getEntries())) {
                            String name = entry.getName();
                            zipOutputStream.putNextEntry(new ZipEntry(name));
                            String override = overrideByEntryName.get(name);
                            if (override != null) {
                                entry = vestigeEarJar.getEntry("META-INF/exploded-assembly-files/" + override);
                            }
                            try (InputStream open = entry.open()) {
                                IOUtils.copy(open, zipOutputStream);
                            }
                            zipOutputStream.closeEntry();
                        }
                    }

                } else {
                    try (InputStream open = vestigeJar.open(); FileOutputStream outputStream = new FileOutputStream(outputFile)) {
                        IOUtils.copy(open, outputStream);
                    }
                }

            }
            return true;
        } else if (file.getName().endsWith(".vwar")) {
            final VestigeWar vestigeWar = VestigeWar.create(file, file.getName().substring(0, file.getName().length() - ".vwar".length()));
            VESTIGE_WAR_THREAD_LOCAL.set(vestigeWar);
            VestigeJar vestigeWarJar = vestigeWar.getVestigeWar();
            try {
                unzipExploded(vestigeWarJar, destinationDir);
            } catch (final IOException e) {
                // If something went wrong, delete extracted dir to keep things clean
                Files.delete(destinationDir);
                throw e;
            }
            VestigeJarEntry explodedAssemblyEntry = vestigeWarJar.getEntry("META-INF/exploded-assembly.properties");
            Map<String, String> entryNameByURL = new HashMap<String, String>();
            if (explodedAssemblyEntry != null) {
                Properties properties = new Properties();
                properties.load(explodedAssemblyEntry.open());
                for (String dep : properties.getProperty("dependencies").split(",")) {
                    entryNameByURL.put(properties.getProperty(dep + ".url"), properties.getProperty(dep + ".name"));
                }
            }

            List<? extends VestigeJar> vestigeWarDependencies = vestigeWar.getVestigeWarDependencies();
            for (VestigeJar vestigeJar : vestigeWarDependencies) {
                String entryName = entryNameByURL.get(vestigeJar.getCodeBase().toString()).replace('/', File.separatorChar);
                File outputFile = new File(destinationDir, entryName);
                try (InputStream open = vestigeJar.open(); FileOutputStream outputStream = new FileOutputStream(outputFile)) {
                    IOUtils.copy(open, outputStream);
                }
            }

            return true;
        }
        return false;
    }

    public File getUnpackDir(final File file) {
        if (file.getName().endsWith(".vear") || file.getName().endsWith(".vwar")) {
            File unpackDir;
            String confAbsolutePath = conf.getAbsolutePath();
            String fileAbsolutePath = file.getParentFile().getAbsolutePath();
            if (fileAbsolutePath.startsWith(confAbsolutePath)) {
                unpackDir = new File(data, fileAbsolutePath.substring(confAbsolutePath.length()));
            } else {
                unpackDir = file.getParentFile();
            }
            return unpackDir;
        }
        return null;
    }

    public WebModuleCreatorData createWebModule(final File warFile, final ClassLoader parentClassLoader) throws OpenEJBException {
        List<? extends VestigeJar> vestigeWarDependencies;
        WebModuleCreatorData webModuleCreatorData = null;
        if (warFile.getName().endsWith(".vwar")) {
            webModuleCreatorData = new WebModuleCreatorData();
            try {
                VestigeWar vestigeWar = VESTIGE_WAR_THREAD_LOCAL.get();
                if (vestigeWar == null) {
                    // already unpacked
                    vestigeWar = VestigeWar.create(warFile, warFile.getName().substring(0, warFile.getName().length() - ".vwar".length()));
                    VESTIGE_WAR_THREAD_LOCAL.set(vestigeWar);
                }

                AttachedClassLoader sharedAttachedClassLoader = vestigeWar.getSharedAttachedClassLoader();
                if (sharedAttachedClassLoader != null) {
                    // parentFirst seems ignored by TomEE, set to false
                    webModuleCreatorData.parentClassLoader = new DelegateClassLoader(parentClassLoader, sharedAttachedClassLoader.getAttachableClassLoader().getClassLoader(),
                            false);
                } else {
                    webModuleCreatorData.parentClassLoader = parentClassLoader;
                }

                webModuleCreatorData.warFile = DeploymentLoader.unpack(warFile);

                final Set<URL> webClassPath = new HashSet<URL>();
                final Set<URL> webRars = new HashSet<URL>();

                webClassPath.add(new File(new File(webModuleCreatorData.warFile, "WEB-INF"), "classes").toURI().toURL());

                vestigeWarDependencies = vestigeWar.getVestigeWarDependencies();
                for (VestigeJar vestigeJar : vestigeWarDependencies) {
                    if (vestigeJar.getName().endsWith(".rar")) {
                        webRars.add(vestigeJar.getFile().toURI().toURL());
                    } else {
                        webClassPath.add(vestigeJar.getFile().toURI().toURL());
                    }
                }

                final WebAppEnricher enricher = SystemInstance.get().getComponent(WebAppEnricher.class);
                if (enricher != null) {
                    webClassPath.addAll(Arrays.asList(enricher.enrichment(null)));
                }

                webModuleCreatorData.urls = new HashMap<String, URL[]>();
                webModuleCreatorData.urls.put(DeploymentLoader.URLS_KEY, webClassPath.toArray(new URL[webClassPath.size()]));
                webModuleCreatorData.urls.put(DeploymentLoader.RAR_URLS_KEY, webRars.toArray(new URL[webRars.size()]));
            } catch (Exception e) {
                throw new OpenEJBException(e);
            }
        } else {
            return null;
        }
        return webModuleCreatorData;
    }

    public void unsetThreadWar() {
        VESTIGE_WAR_THREAD_LOCAL.remove();
    }

    public AppModuleCreatorData createAppModule(final File jarFile, final ClassLoader parent) throws OpenEJBException {
        if (!jarFile.getName().endsWith(".vear")) {
            return null;
        }

        AppModuleCreatorData appModuleCreatorData = new AppModuleCreatorData();
        try {
            VestigeEar vestigeEar = VESTIGE_EAR_THREAD_LOCAL.get();
            if (vestigeEar == null) {
                // already unpacked
                vestigeEar = VestigeEar.create(jarFile, jarFile.getName().substring(0, jarFile.getName().length() - ".vear".length()));
                VESTIGE_EAR_THREAD_LOCAL.set(vestigeEar);
            }

            final VestigeJar vestigeEarJar = vestigeEar.getVestigeEar();

            URL baseURL = new URL("jar", "", -1, vestigeEarJar.getFile().toURI().getPath() + "!/");
            VestigeJarEntry appXML = vestigeEarJar.getEntry("META-INF/application.xml");
            if (appXML != null) {
                appModuleCreatorData.application = ApplicationXml.unmarshal(appXML.open());
                appModuleCreatorData.applicationFound = true;
            } else {
                appModuleCreatorData.application = new Application();
            }

            appModuleCreatorData.appDescriptors = new HashMap<String, URL>();

            for (String descriptor : DeploymentLoader.KNOWN_DESCRIPTORS) {
                VestigeJarEntry entry = vestigeEarJar.getEntry("META-INF/" + descriptor);
                if (entry != null) {
                    String name = entry.getName();
                    URL url = new URL(baseURL, name);
                    appModuleCreatorData.appDescriptors.put(name, url);
                }
            }

            File file = vestigeEarJar.getFile();
            appModuleCreatorData.appUrl = file.toURI().toURL();
            appModuleCreatorData.appId = file.getAbsolutePath();
            appModuleCreatorData.appDir = DeploymentLoader.unpack(jarFile);

            AttachedClassLoader sharedAttachedClassLoader = vestigeEar.getSharedAttachedClassLoader();
            if (sharedAttachedClassLoader != null) {
                appModuleCreatorData.parentClassLoader = new DelegateClassLoader(parent, sharedAttachedClassLoader.getAttachableClassLoader().getClassLoader(), true);
            } else {
                appModuleCreatorData.parentClassLoader = parent;
            }

            appModuleCreatorData.finder = new ResourceFinder() {

                @Override
                public URL find(final String uri) throws IOException {
                    String url = "jar:" + vestigeEarJar.getCodeBase().toString() + "!/" + uri;
                    VestigeJarEntry entry = vestigeEarJar.getEntry(uri);
                    try {
                        return new URL(null, url, new URLStreamHandler() {

                            @Override
                            protected URLConnection openConnection(final URL u) throws IOException {
                                return new VestigeJarEntryConnection(u, null, entry);
                            }
                        });
                    } catch (MalformedURLException e) {
                        return null;
                    }
                }
            };

        } catch (Exception e) {
            throw new OpenEJBException(e);
        }
        return appModuleCreatorData;
    }

    @Override
    public Policy getPolicy() {
        return tomEEVestigeSystem.getPolicy();
    }

    @Override
    public void setPolicy(final Policy policy) {
        tomEEVestigeSystem.setPolicy(policy);
    }

    @Override
    public File toFile(final URL url) {
        String protocol = url.getProtocol();
        if (VestigeCoreURLStreamHandler.PROTOCOL.equals(protocol)) {
            Matcher matcher = VestigeCoreURLStreamHandler.URL_PATH_PATTERN.matcher(url.getPath());
            if (!matcher.matches()) {
                throw new IllegalArgumentException("Invalid URL: " + url.toExternalForm());
            }
            try {
                return new File(URLDecoder.decode(matcher.group(VestigeCoreURLStreamHandler.URL_PATH_PATTERN_FILE_NAME_POSITION), "UTF-8"));
            } catch (UnsupportedEncodingException e) {
                throw new IllegalArgumentException(e);
            }
        }
        String urlString = url.toString();
        if (urlString.startsWith("jar:mvn:")) {
            try {
                URLConnection connection = url.openConnection();
                if (connection instanceof VestigeJarEntryConnection) {
                    VestigeJarEntryConnection vestigeJarEntryConnection = (VestigeJarEntryConnection) connection;
                    VestigeJar entryVestigeJar = vestigeJarEntryConnection.getEntryVestigeJar();
                    if (entryVestigeJar != null) {
                        return entryVestigeJar.getFile();
                    }
                    return vestigeJarEntryConnection.getVestigeJar().getFile();
                }
            } catch (IOException e) {
            }
        }

        return null;
    }

    public boolean filterApplication(final String name) {
        return name.matches(".*\\.(jar|war|rar|ear|vwar|vear)");
    }

    @Override
    public ClassLoader createSharedLoader(final ClassLoader parent, final boolean parentFirst) {
        VestigeWar vestigeWar = VESTIGE_WAR_THREAD_LOCAL.get();
        if (vestigeWar == null) {
            return parent;
        }
        AttachedClassLoader sharedAttachedClassLoader = vestigeWar.getSharedAttachedClassLoader();
        if (sharedAttachedClassLoader == null) {
            return parent;
        }
        ClassLoader sharedClassLoader = sharedAttachedClassLoader.getAttachableClassLoader().getClassLoader();
        return new DelegateClassLoader(parent, sharedClassLoader, false);
    }

    public ClassLoader createAppSharedLoader(final ClassLoader parent) {
        VestigeEar vestigeEar = VESTIGE_EAR_THREAD_LOCAL.get();
        if (vestigeEar == null) {
            return parent;
        }
        AttachedClassLoader sharedAttachedClassLoader = vestigeEar.getSharedAttachedClassLoader();
        if (sharedAttachedClassLoader == null) {
            return parent;
        }
        ClassLoader sharedClassLoader = sharedAttachedClassLoader.getAttachableClassLoader().getClassLoader();
        return new DelegateClassLoader(parent, sharedClassLoader, true);
    }

    public void associateWithThreadWar(final Object toAssociate) {
        map.put(toAssociate, VESTIGE_WAR_THREAD_LOCAL.get());
    }

    public void associateWithWar(final Object alreadyAssociated, final Object toAssociate) {
        map.put(toAssociate, map.get(alreadyAssociated));
    }

    public void associateCurrentThreadWithWar(final Object tomEEWebappLoader) {
        VestigeWar vestigeWar = map.get(tomEEWebappLoader);
        VESTIGE_WAR_THREAD_LOCAL.set(vestigeWar);
    }

}
